/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file http_recv.c paraslash's http receiver */

#include "para.h"

#include "http.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "http_recv.cmdline.h"
#include "error.h"
#include "net.h"
#include "string.h"
#include "fd.h"

/** the output buffer size of the http receiver */
#define BUFSIZE (32 * 1024)

/**
 * the possible states of a http receiver node
 *
 * \sa receiver_node
 */
enum http_recv_status {HTTP_CONNECTED, HTTP_SENT_GET_REQUEST, HTTP_STREAMING};

/**
 * data specific to the http receiver
 *
 * Each running instance of the http receiver reserves space for one such struct.
 */
struct private_http_recv_data {
/**
 *
 *
 * the current status of the http receiver node
 *
 * It gets initialized to #HTTP_CONNECTED by the open function of the
 * http receiver.
 *
 * \sa receiver::open, receiver_node.
 */
	enum http_recv_status status;
/**
 *
 *
 * the file descriptor used for receiving the http stream
 *
 * The pre_select function of the http receiver adds this file descriptor to
 * the set of file decriptors which are checked for reading/writing (depending
 * on the current status) by the select loop of the application (para_audiod or
 * para_recv).
 *
 * The post_select function of the http receiver uses \a fd, if ready, to
 * establish the http connection, and updates \a status according to the new
 * state of the connection.  As soon as \a status is #HTTP_STREAMING, \a fd is
 * going to be only checked for reading. If data is available, it is read into
 * the output buffer of the receiver node by post_select.
 *
 * \sa receiver::pre_select receiver::post_select receiver_node
 */
	int fd;
};

static void http_shutdown(void)
{
	return;
}

static char *make_request_msg(void)
{
	char *ret, *hn = para_hostname();
	ret = make_message("%s1.0\nHost: %s\nUser-Agent: para_recv/%s\n\n\n",
		HTTP_GET_MSG, hn, PACKAGE_VERSION);
	free(hn);
	return ret;
}

static void http_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_http_recv_data *phd = rn->private_data;

	t->ret = 1;
	if  (phd->status == HTTP_CONNECTED)
		para_fd_set(phd->fd, &s->wfds, &s->max_fileno);
	else
		para_fd_set(phd->fd, &s->rfds, &s->max_fileno);
}


static void http_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_http_recv_data *phd = rn->private_data;

	t->ret = -E_HTTP_RECV_EOF;
	if (rn->output_eof && *rn->output_eof)
		goto out;
	t->ret = 1;
	if (!s->select_ret)
		goto out;
	if  (phd->status == HTTP_CONNECTED) {
		char *rq;
		if (!FD_ISSET(phd->fd, &s->wfds))
			goto out;
		rq = make_request_msg();
		PARA_INFO_LOG("%s", "sending http request\n");
		t->ret = send_va_buffer(phd->fd, "%s", rq);
		free(rq);
		if (t->ret > 0)
			phd->status = HTTP_SENT_GET_REQUEST;
		goto out;
	}
	if (!FD_ISSET(phd->fd, &s->rfds))
		goto out;
	if (phd->status == HTTP_SENT_GET_REQUEST) {
		t->ret = recv_pattern(phd->fd, HTTP_OK_MSG, MAXLINE);
		if (t->ret < 0)
			goto out;
		PARA_INFO_LOG("%s", "received ok msg, streaming\n");
		t->ret = 1;
		phd->status = HTTP_STREAMING;
		goto out;
	}
	t->ret = -E_OVERRUN;
	if (rn->loaded >= BUFSIZE)
		goto out;
	t->ret = recv_bin_buffer(phd->fd, rn->buf + rn->loaded,
		BUFSIZE - rn->loaded);
	if (t->ret <= 0) {
		if (!t->ret)
			t->ret = -E_HTTP_RECV_EOF;
		goto out;
	}
	rn->loaded += t->ret;
out:
	if (t->ret < 0)
		rn->eof = 1;
}

static void http_recv_close(struct receiver_node *rn)
{
	struct private_http_recv_data *phd = rn->private_data;
	close(phd->fd);
	free(rn->buf);
	free(rn->private_data);
}

static void *http_recv_parse_config(int argc, char **argv)
{
	struct http_recv_args_info *tmp = para_calloc(sizeof(struct http_recv_args_info));

	if (!http_recv_cmdline_parser(argc, argv, tmp))
		return tmp;
	free(tmp);
	return NULL;
}

static int http_recv_open(struct receiver_node *rn)
{
	struct private_http_recv_data *phd;
	struct hostent *he;
	struct http_recv_args_info *conf = rn->conf;
	struct sockaddr_in their_addr;
	int ret;

	rn->buf = para_calloc(BUFSIZE);
	rn->private_data = para_calloc(sizeof(struct private_http_recv_data));
	phd = rn->private_data;
	ret = get_host_info(conf->host_arg, &he);
	if (!ret < 0)
		goto err_out;
	/* get new socket */
	ret = get_socket();
	if (ret < 0)
		goto err_out;
	phd->fd = ret;
	/* init their_addr */
	init_sockaddr(&their_addr, conf->port_arg, he);
	/* connect */
	PARA_NOTICE_LOG("connecting to %s:%d\n", conf->host_arg,
		conf->port_arg);
	ret = para_connect(phd->fd, &their_addr);
	if (ret < 0) {
		close(phd->fd);
		goto err_out;
	}
	mark_fd_nonblock(phd->fd);
	phd->status = HTTP_CONNECTED;
	return 1;
err_out:
	free(rn->private_data);
	free(rn->buf);
	return ret;
}

/**
 * the init function of the http receiver
 *
 * \param r pointer to the receiver struct to initialize
 *
 * Just initialize all function pointers of \a r.
 */
void http_recv_init(struct receiver *r)
{
	r->open = http_recv_open;
	r->close = http_recv_close;
	r->pre_select = http_recv_pre_select;
	r->post_select = http_recv_post_select;
	r->shutdown = http_shutdown;
	r->parse_config = http_recv_parse_config;
}
