/*
 * Copyright (C) 2004-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/** \file signal.c signal handling functions */



#include "para.h"
#include "fd.h"

#include <signal.h>

#include "error.h"
static int signal_pipe[2];

/**
 * initialize the paraslash signal subsystem
 *
 * This function creates a pipe, the signal pipe, to deliver pending
 * signals to the application (Bernstein's trick). It should be called
 * during the application's startup part, followed by subsequent calls
 * to para_install_sighandler() for each signal that should be caught.
 *
 * para_signal_init() installs a generic signal handler which is used for all
 * signals simultaneously. When a signal arrives, this generic signal handler
 * writes the corresponding signal number to the signal pipe so that the
 * application can test for pending signals simply by checking the signal pipe
 * for reading, e.g. by using the select(2) system call.
 *
 * \return This function either succeeds or calls exit(2) to terminate
 * the current process. On success, the file descriptor of the signal pipe is
 * returned.
 */
int para_signal_init(void)
{
	int ret = -E_SIGNAL_PIPE;
	if (pipe(signal_pipe))
		goto err_out;
	ret = mark_fd_nonblock(signal_pipe[0]);
	if (ret < 0)
		goto err_out;
	ret = mark_fd_nonblock(signal_pipe[1]);
	if (ret < 0)
		goto err_out;
	return signal_pipe[0];
err_out:
	PARA_EMERG_LOG("%s\n", PARA_STRERROR(-ret));
	exit(EXIT_FAILURE);
}

/*
 * just write one integer to signal pipe
 */
static void generic_signal_handler(int s)
{
	write(signal_pipe[1], &s, sizeof(int));
	//fprintf(stderr, "got sig %i, write returned %d\n", s, ret);
}

/**
 * reap one child
 *
 * call waitpid() and print a log message containing the pid
 * and the cause of the child's death.
 *
 * \return Like \p waitpid(), this function returns the process ID of the
 * terminated child; on error, \p -E_WAITPID is returned.
 * \sa waitpid(2)
 */
pid_t para_reap_child(void)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);

	if (pid <= 0) {
		if (pid < 0)
			pid = -E_WAITPID;
		return pid;
	}
	if (WIFEXITED(status))
		PARA_DEBUG_LOG("child %i exited. Exit status: %i\n", pid,
			WEXITSTATUS(status));
	else if (WIFSIGNALED(status))
		PARA_DEBUG_LOG("child %i was killed by signal %i\n", pid,
			WTERMSIG(status));
	else
		PARA_WARNING_LOG("child %i terminated abormally\n", pid);
	return pid;
}

/**
 * paraslash's zombie killer
 *
 * It just calls \p para_reap_child() until there are no more children left to
 * reap.
 */
void para_reap_children(void)
{
	while (para_reap_child() > 0)
		; /* nothing */
}

/**
 * wrapper around signal(2)
 * \param sig the number of the signal to catch
 *
 * This installs the generic signal handler for the given signal.
 * \return This function returns 1 on success and \p -E_SIGNAL_SIG_ERR on errors.
 * \sa signal(2)
 */
int para_install_sighandler(int sig)
{
	PARA_DEBUG_LOG("catching signal %d\n", sig);
	return signal(sig, &generic_signal_handler) == SIG_ERR?  -E_SIGNAL_SIG_ERR : 1;
}

/**
 * return number of next pending signal
 *
 * This should be called if the fd for the signal pipe is ready for reading.
 *
 * \return On success, the number of the received signal is returned. \p
 * -E_SIGNAL_READ is returned if a read error occured while reading the signal
 * pipe.  If the read was interrupted by another signal the function returns 0.
 */
int para_next_signal(void)
{
	int s;
	ssize_t r;

	if ((r = read(signal_pipe[0], &s, sizeof(s)) == sizeof(s)) > 0) {
		PARA_DEBUG_LOG("next signal: %d\n", s);
		return s;
	}
	return r < 0 && (errno != EAGAIN)? 0 : -E_SIGNAL_READ;
}
